/****************************************************************************/
/*                   Canon PowerShot firmware font editor                   */
/****************************************************************************/
/* Created: 5. 9. 2004 [DJGPP]                                              */
/* Last modified: 5. 9. 2004 [DJGPP]                                        */
/* Copyright (C) 2004 by RayeR/Martin Rehak                                 */
/* Contact: rayer@seznam.cz, rehakm2@feld.cvut.cz                           */
/****************************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <dos.h>
#include <io.h>

#define FONT_VISIBLE_WIDTH 22          // pocet viditelnych bodu fontu na radek 22/44
#define FONT_STORAGE_WIDTH 24          // pocet bitu na radek ve fontfajlu 24/48
#define FONT_HEIGHT 24                 // pocet viditelnych bodu fontu na sloupec
#define FONT_OFFSET 32                 // ASCII kod pocatecniho znaku fontu
typedef struct {                       // struktura radku fontu, obvykle 3B/radek
  unsigned long long fontline : (FONT_STORAGE_WIDTH) __attribute__((packed));
} FONTLINE;

FONTLINE fontchar[FONT_HEIGHT];        // bitove pole jednoho znaku fontu
FONTLINE fontchar_clipbrd[FONT_HEIGHT];// bitove pole jednoho znaku fontu pro schranku
char disp0='.', disp1='X';             // znak kterym se zobrazi 0 a 1
FILE *fontfile;                        // soubor s fontem
char key;                              // cteni klavesy
int flip=0, charno=0, clipptr=-1;      // flag otoceni, pocitadlo, pozice schranky
int cursorx=0, cursory=0;              // souradnice kurzoru

//***************** zobrazi radek bitove mapy fontu s danym poctem bitu/radek
void display_font_line(long long bitstream, int bits, int flipped)
{                                      // flipped=0 : LSB...MSB
  int i,j;                             // flipped=1 : MSB...LSB
  for (i=0; i<bits; i++)
    {
    if (flipped>0)
      j=bits-i-1;
    else
      j=i;
    if (((bitstream>>j)&0x1)>0)
      putchar(disp1);
    else
      putchar(disp0);
    }
  putchar('\n');
  return;
}

//***************** zobrazi jeden znak fontu v textovem rezimu na obrazovce
void display_font_char(FONTLINE *fontchar, int flipped)
{
  int i;
  for (i=0; i<FONT_HEIGHT; i++)
    display_font_line(fontchar[i].fontline,FONT_VISIBLE_WIDTH,flipped);
  return;
}

int main(int argc, char *argv[])
{
  printf("\nCanon PowerShot firmware font editor 1.0 (C) 2004 by RayeR; rayer@seznam.cz\n");
  printf("Compiled by GCC %s at %s, %s\n\n", __VERSION__, __TIME__, __DATE__);
  if (argc<2)                          // pokud na radce neni zadny parametr
    {
    printf("SYNTAX:   fontedit fontfile\n");
    printf("          where fontfile is canon font file 22x24\n");
    printf("CONTROLS: ESC - exit\n");
    printf("          PG-UP/DOWN - previous/next character\n");
    printf("          UP/DOWN/LEFT/RIGHT/HOME/END/-/+ - cursor moving\n");
    printf("          SPACE - toggle font bit\n");
    printf("          C - clear font line\n");
    printf("          DEL/INS - copy/paste character to/from clipboard\n");
    printf("          ENTER - store current character to file\n");
    printf("          F - horizontal font flip (display only)\n");
    printf("          default odrer is LSB...MSB (left to right)\n");
    exit(1);
  }

  if ((fontfile=fopen(argv[1],"rb+"))==NULL)
    {                                  // otevri soubor fontu pro cteni a zapis
    printf("FONTEDIT ERROR: cannot open font file %s\n",argv[1]);
    exit(255);
    }

  delay(1000);
  fread(fontchar,sizeof(fontchar),1,fontfile);
  memset(fontchar_clipbrd,0,sizeof(fontchar_clipbrd));
  do                                   // hlavni smycka programu
    {
    if ((key=='I')&&((sizeof(fontchar)*(charno+1))<filelength(fileno(fontfile)))) // char+
      {
      charno++;
      fseek(fontfile,sizeof(fontchar)*charno,SEEK_SET);
      fread(fontchar,sizeof(fontchar),1,fontfile);
      }
    if ((key=='Q')&&(charno>0))        // char-
      {
      charno--;
      fseek(fontfile,sizeof(fontchar)*charno,SEEK_SET);
      fread(fontchar,sizeof(fontchar),1,fontfile);
      }
    if ((key=='f')||(key=='F'))        // flip toggle
      {
      if (flip!=0)
        flip=0;
      else
        flip=1;
      }
    if ((key=='P')&&(wherey()<FONT_HEIGHT)) // kurzor dolu
      cursory++;
    if ((key=='H')&&(wherey()>1))      // kurzor nahoru
      cursory--;
    if ((key=='M')&&(wherex()<FONT_VISIBLE_WIDTH)) // kurzor doprava
      cursorx++;
    if ((key=='K')&&(wherex()>1))      // kurzor doleva
      cursorx--;
    if (key=='G')                      // kurzor na zacatek
      cursorx=0;
    if (key=='O')                      // kurzor na konec
      cursorx=FONT_VISIBLE_WIDTH-1;
    if (key=='-')                      // kurzor na prvni radek
      cursory=0;
    if (key=='+')                      // kurzor na posledni radek
      cursory=FONT_HEIGHT-1;
    if (key==' ')                      // toggle font bit
      fontchar[cursory].fontline=fontchar[cursory].fontline^(1<<cursorx);
    if ((key=='c')||(key=='C'))        // clear font line
      fontchar[cursory].fontline=0;
    if (key=='S')                      // kopirovani znaku do schranky
      {
      memcpy(fontchar_clipbrd,fontchar,sizeof(fontchar));
      clipptr=charno;
      }
    if (key=='R')                      // kopirovani znaku ze schranky
      memcpy(fontchar,fontchar_clipbrd,sizeof(fontchar));
    if (key==13)                       // uloz editovany znak do souboru
      {
      fseek(fontfile,sizeof(fontchar)*charno,SEEK_SET);
      if (fwrite(fontchar,sizeof(fontchar),1,fontfile)==0)
        {
        printf("\nFONTEDIT ERROR: font file %s is write-protected\n",argv[1]);
        delay(2000);
        }
      }
    clrscr();
    display_font_char(fontchar,flip);  // zobraz znak
    printf("character #%d, ASCII code: %d (%02Xh); bit coordinates: [%d,%d]",charno,charno+FONT_OFFSET,charno+FONT_OFFSET,cursorx,cursory);
    if (clipptr>=0)                    // clipboard pointer
      printf(", clipbrd=%d",clipptr);
    if (flip!=0)                       // flip flag
      printf(", F");
    gotoxy(cursorx+1,cursory+1);       // nastav kurzor podle souradnic
    } while ((key=getch())!=27);       // exit on ESC
  fclose(fontfile);                    // zavri soubor
  clrscr();
  return(0);
}
